#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
import os
import gzip
import urllib.request, urllib.parse, urllib.error
from ..command import Command
from ..tools import process, question
from ..phpunit import PHPUnit


class PhpunitCommand(Command):

    _arguments = [
        (
            ['-f', '--force'],
            {
                'action': 'store_true',
                'help': 'force the initialisation'
            }
        ),
        (
            ['-r', '--run'],
            {
                'action': 'store_true',
                'help': 'also run the tests'
            }
        ),
        (
            ['-t', '--testcase'],
            {
                'default': None,
                'help': 'testcase class to run (From Moodle 2.6)',
                'metavar': 'testcase'
            }
        ),
        (
            ['-s', '--testsuite'],
            {
                'default': None,
                'help': 'testsuite to run',
                'metavar': 'testsuite'
            }
        ),
        (
            ['-u', '--unittest'],
            {
                'default': None,
                'help': 'test file to run',
                'metavar': 'path'
            }
        ),
        (
            ['-k', '--skip-init'],
            {
                'action': 'store_true',
                'dest': 'skipinit',
                'help': 'allows tests to start quicker when the instance is already initialised'
            }
        ),
        (
            ['-q', '--stop-on-failure'],
            {
                'action': 'store_true',
                'dest': 'stoponfailure',
                'help': 'stop execution upon first failure or error'
            }
        ),
        (
            ['-c', '--coverage'],
            {
                'action': 'store_true',
                'help': 'creates the HTML code coverage report'
            }
        ),
        (
            ['--filter'],
            {
                'default': None,
                'help': 'filter to pass through to PHPUnit',
                'metavar': 'filter'
            }
        ),
        (
            ['--repeat'],
            {
                'default': None,
                'help': 'run tests repeatedly for the given number of times',
                'metavar': 'times',
                'type': int
            }
        ),
        (
            ['name'],
            {
                'default': None,
                'help': 'name of the instance',
                'metavar': 'name',
                'nargs': '?'
            }
        )
    ]
    _description = 'Initialize PHPUnit'

    def run(self, args):

        # Loading instance
        M = self.Wp.resolve(args.name)
        if not M:
            raise Exception('This is not a Moodle instance')

        # Check if installed
        if not M.get('installed'):
            raise Exception('This instance needs to be installed first')

        # Check if testcase option is available.
        if args.testcase and M.branch_compare('26', '<'):
            self.argumentError('The --testcase option only works with Moodle 2.6 or greater.')

        # Create the Unit test object.
        PU = PHPUnit(self.Wp, M)

        # Skip init.
        if not args.skipinit:
            self.init(M, PU, args)

        # Automatically add the suffix _testsuite.
        testsuite = args.testsuite
        if testsuite and not testsuite.endswith('_testsuite'):
            testsuite += '_testsuite'

        # Check repeat arg.
        repeat = args.repeat
        if repeat:
            # Make sure repeat is greater than 1.
            if repeat <= 1:
                repeat = None

        kwargs = {
            'coverage': args.coverage,
            'filter': args.filter,
            'testcase': args.testcase,
            'testsuite': testsuite,
            'unittest': args.unittest,
            'stopon': [] if not args.stoponfailure else ['failure'],
            'repeat': repeat
        }

        if args.run:
            PU.run(**kwargs)
            if args.coverage:
                logging.info('Code coverage is available at: \n %s', (PU.getCoverageUrl()))
        else:
            logging.info('Start PHPUnit:\n %s', (' '.join(PU.getCommand(**kwargs))))

    def init(self, M, PU, args):
        """Initialises PHP Unit"""

        # Install Composer
        if PU.usesComposer():
            if not os.path.isfile(os.path.join(M.get('path'), 'composer.phar')):
                logging.info('Installing Composer')
                cliFile = 'phpunit_install_composer.php'
                cliPath = os.path.join(M.get('path'), 'phpunit_install_composer.php')
                (to, headers) = urllib.request.urlretrieve('http://getcomposer.org/installer', cliPath)
                if headers.get('content-encoding') == 'gzip':
                    f = gzip.open(cliPath, 'r')
                    content = f.read()
                    f.close()
                    f = open(cliPath, 'w')
                    f.write(content)
                    f.close()
                M.cli('/' + cliFile, stdout=None, stderr=None)
                os.remove(cliPath)
                M.cli('composer.phar', args='install', stdout=None, stderr=None)

        # If Oracle, ask the user for a Behat prefix, if not set.
        prefix = M.get('phpunit_prefix')
        if M.get('dbtype') == 'oci' and (args.force or not prefix or len(prefix) > 2):
            while not prefix or len(prefix) > 2:
                prefix = question('What prefix would you like to use? (Oracle, max 2 chars)')
        else:
            prefix = None

        PU.init(force=args.force, prefix=prefix)





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import re
import logging

from ..db import DB
from ..command import Command
from ..tools import yesOrNo, version_options
from ..exceptions import CreateException, InstallException


class CreateCommand(Command):

    _description = 'Creates new instances of Moodle'

    def __init__(self, *args, **kwargs):
        super(CreateCommand, self).__init__(*args, **kwargs)
        self._arguments = [
            (
                ['-i', '--install'],
                {
                    'action': 'store_true',
                    'dest': 'install',
                    'help': 'launch the installation script after creating the instance'
                }
            ),
            (
                ['-e', '--engine'],
                {
                    'action': 'store',
                    'choices': ['mariadb', 'mysqli', 'pgsql', 'sqlsrv'],
                    'default': self.C.get('defaultEngine'),
                    'help': 'database engine to install the instance on, use with --install',
                    'metavar': 'engine'
                }
            ),
            (
                ['-t', '--integration'],
                {
                    'action': 'store_true',
                    'help': 'create an instance from integration'
                }
            ),
            (
                ['-r', '--run'],
                {
                    'action': 'store',
                    'help': 'scripts to run after installation',
                    'metavar': 'run',
                    'nargs': '*'
                }
            ),
            (
                ['-n', '--identifier'],
                {
                    'action': 'store',
                    'default': None,
                    'help': 'use this identifier instead of generating one. The flag --suffix will be used. ' +
                        'Do not use when creating multiple versions at once',
                    'metavar': 'name',
                }
            ),
            (
                ['-s', '--suffix'],
                {
                    'action': 'store',
                    'default': [None],
                    'help': 'suffixes for the instance name',
                    'metavar': 'suffix',
                    'nargs': '*'
                }
            ),
            (
                ['-v', '--version'],
                {
                    'choices': version_options(),
                    'default': ['main'],
                    'help': 'version of Moodle',
                    'metavar': 'version',
                    'nargs': '*'
                }
            ),
        ]

    def run(self, args):

        engine = args.engine
        versions = args.version
        suffixes = args.suffix
        install = args.install

        # Throw an error when --engine is used without --install. The code is currently commented out
        # because as --engine has a default value, it will always be set, and so it becomes impossible
        # to create an instance without installing it. I cannot think about a clean fix yet. Removing
        # the default value will cause --help not to output the default as it should... Let's put more
        # thoughts into this and perhaps use argument groups.
        # if engine and not install:
            # self.argumentError('--engine can only be used with --install.')

        for version in versions:
            for suffix in suffixes:
                arguments = {
                    'version': version,
                    'suffix': suffix,
                    'engine': engine,
                    'integration': args.integration,
                    'identifier': args.identifier,
                    'install': install,
                    'run': args.run
                }
                self.do(arguments)
                logging.info('')

        logging.info('Process complete!')

    def do(self, args):
        """Proceeds to the creation of an instance"""

        # TODO Remove these ugly lines, but I'm lazy to rewrite the variables in this method...
        class Bunch:
            __init__ = lambda self, **kw: setattr(self, '__dict__', kw)
        args = Bunch(**args)

        engine = args.engine
        version = args.version
        name = self.Wp.generateInstanceName(version, integration=args.integration, suffix=args.suffix, identifier=args.identifier)

        # Wording version
        versionNice = version
        if version in ['master', 'main']:
            versionNice = self.C.get('wording.master')

        # Generating names
        if args.integration:
            fullname = self.C.get('wording.integration') + ' ' + versionNice + ' ' + self.C.get('wording.%s' % engine)
        else:
            fullname = self.C.get('wording.stable') + ' ' + versionNice + ' ' + self.C.get('wording.%s' % engine)

        # Append the suffix
        if args.suffix:
            fullname += ' ' + args.suffix.replace('-', ' ').replace('_', ' ').title()

        # Create the instance
        logging.info('Creating instance %s...' % name)
        kwargs = {
            'name': name,
            'version': version,
            'integration': args.integration
        }
        try:
            M = self.Wp.create(**kwargs)
        except CreateException as e:
            logging.error('Error creating %s:\n  %s' % (name, e))
            return False
        except Exception as e:
            logging.exception('Error creating %s:\n  %s' % (name, e))
            return False

        # Run the install script
        if args.install:

            # Checking database
            dbname = re.sub(r'[^a-zA-Z0-9]', '', name).lower()
            prefixDbname = self.C.get('db.namePrefix')
            if prefixDbname:
                dbname = prefixDbname + dbname
            dbname = dbname[:28]
            db = DB(engine, self.C.get('db.%s' % engine))
            dropDb = False
            if db.dbexists(dbname):
                logging.info('Database already exists (%s)' % dbname)
                dropDb = yesOrNo('Do you want to remove it?')

            # Install
            kwargs = {
                'engine': engine,
                'dbname': dbname,
                'dropDb': dropDb,
                'fullname': fullname,
                'dataDir': self.Wp.getPath(name, 'data'),
                'wwwroot': self.Wp.getUrl(name)
            }
            try:
                M.install(**kwargs)
            except InstallException as e:
                logging.warning('Error while installing %s:\n  %s' % (name, e))
                return False
            except Exception as e:
                logging.exception('Error while installing %s:\n  %s' % (name, e))
                return False

            # Running scripts
            if M.isInstalled() and type(args.run) == list:
                for script in args.run:
                    logging.info('Running script \'%s\'' % (script))
                    try:
                        M.runScript(script)
                    except Exception as e:
                        logging.warning('Error while running the script \'%s\':\  %s' % (script, e))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
from ..command import Command


class InfoCommand(Command):

    _arguments = [
        (
            ['-e', '--edit'],
            {
                'dest': 'edit',
                'help': 'value to set to the variable (--var). This value will be set in the config file of the instance. Prepend the value with i: or b: to set as int or boolean. DO NOT use names used by MDK (identifier, stablebranch, ...).',
                'metavar': 'value',
                'nargs': '?'
            }
        ),
        (
            ['-i', '--integration'],
            {
                'action': 'store_true',
                'dest': 'integration',
                'help': 'used with --list, only display integration instances'
            }
        ),
        (
            ['-l', '--list'],
            {
                'action': 'store_true',
                'dest': 'list',
                'help': 'list the instances'
            }
        ),
        (
            ['-n', '--name-only'],
            {
                'action': 'store_true',
                'dest': 'nameonly',
                'help': 'used with --list, only display instances name'
            }
        ),
        (
            ['-s', '--stable'],
            {
                'action': 'store_true',
                'dest': 'stable',
                'help': 'used with --list, only display stable instances'
            }
        ),
        (
            ['-v', '--var'],
            {
                'default': None,
                'help': 'variable to output or edit',
                'metavar': 'var',
                'nargs': '?'
            }
        ),
        (
            ['name'],
            {
                'default': None,
                'help': 'name of the instance',
                'metavar': 'name',
                'nargs': '?'
            }
        )
    ]
    _description = 'Information about a Moodle instance'

    def run(self, args):
        # List the instances
        if args.list:
            if args.integration != False or args.stable != False:
                l = self.Wp.list(integration=args.integration, stable=args.stable)
            else:
                l = self.Wp.list()
            l.sort()
            for i in l:
                if not args.nameonly:
                    M = self.Wp.get(i)
                    print('{0:<25}'.format(i), M.get('release'))
                else:
                    print(i)

        # Loading instance
        else:
            M = self.Wp.resolve(args.name)
            if not M:
                raise Exception('This is not a Moodle instance')

            # Printing/Editing variable.
            if args.var != None:
                # Edit a value.
                if args.edit != None:
                    val = args.edit
                    if val.startswith('b:'):
                        val = True if val[2:].lower() in ['1', 'true'] else False
                    elif val.startswith('i:'):
                        try:
                            val = int(val[2:])
                        except ValueError:
                            # Not a valid int, let's consider it a string.
                            pass
                    M.updateConfig(args.var, val)
                    logging.info('Set $CFG->%s to %s on %s' % (args.var, str(val), M.get('identifier')))
                else:
                    print(M.get(args.var))

            # Printing info
            else:
                infos = M.info()
                for key in sorted(infos.keys()):
                    print('{0:<20}: {1}'.format(key, infos[key]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

from datetime import datetime
import textwrap
import re
from ..command import Command
from ..jira import Jira
from ..tools import parseBranch, getText


class TrackerCommand(Command):

    _arguments = [
        (
            ['--open'],
            {
                'action': 'store_true',
                'help': 'Open issue in browser'
            }
        ),
        (
            ['-t', '--testing'],
            {
                'action': 'store_true',
                'help': 'include testing instructions'
            }
        ),
        (
            ['issue'],
            {
                'help': 'MDL issue number. Guessed from the current branch if not specified.',
                'nargs': '?'
            }
        ),
        (
            ['--add-labels'],
            {
                'action': 'store',
                'dest':  'addlabels',
                'help': 'add the specified labels to the issue',
                'metavar':  'labels',
                'nargs': '+',
            }
        ),
        (
            ['--remove-labels'],
            {
                'action': 'store',
                'dest':  'removelabels',
                'help': 'remove the specified labels from the issue',
                'metavar':  'labels',
                'nargs': '+',
            }
        ),
        (
            ['--comment'],
            {
                'action': 'store_true',
                'help': 'add a comment to the issue',
            }
        )
    ]
    _description = 'Interact with Moodle tracker'

    Jira = None
    mdl = None

    def run(self, args):

        issue = None
        if not args.issue:
            M = self.Wp.resolve()
            if M:
                parsedbranch = parseBranch(M.currentBranch())
                if parsedbranch:
                    issue = parsedbranch['issue']
        else:
            issue = args.issue

        if not issue or not re.match('(MDL|mdl)?(-|_)?[1-9]+', issue):
            raise Exception('Invalid or unknown issue number')

        self.mdl = 'MDL-' + re.sub(r'(MDL|mdl)(-|_)?', '', issue)

        if args.open:
            Jira.openInBrowser(self.mdl)
            return

        self.Jira = Jira()

        if args.addlabels:
            if 'triaged' in args.addlabels:
                self.argumentError('The label \'triaged\' cannot be added using MDK')
            elif 'triaging_in_progress' in args.addlabels:
                self.argumentError('The label \'triaging_in_progress\' cannot be added using MDK')
            self.Jira.addLabels(self.mdl, args.addlabels)

        if args.removelabels:
            if 'triaged' in args.removelabels:
                self.argumentError('The label \'triaged\' cannot be removed using MDK')
            elif 'triaging_in_progress' in args.removelabels:
                self.argumentError('The label \'triaging_in_progress\' cannot be removed using MDK')
            self.Jira.removeLabels(self.mdl, args.removelabels)

        if args.comment:
            comment = getText()
            self.Jira.addComment(self.mdl, comment)

        self.info(args)

    def info(self, args):
        """Display classic information about an issue"""
        issue = self.Jira.getIssue(self.mdl)

        title = '%s: %s' % (issue['key'], issue['fields']['summary'])
        created = datetime.strftime(Jira.parseDate(issue['fields'].get('created')), '%Y-%m-%d %H:%M')
        resolution = '' if issue['fields']['resolution'] == None else '(%s)' % (issue['fields']['resolution']['name'])
        resolutiondate = ''
        if issue['fields'].get('resolutiondate') != None:
            resolutiondate = datetime.strftime(Jira.parseDate(issue['fields'].get('resolutiondate')), '%Y-%m-%d %H:%M')
        print('-' * 72)
        for l in textwrap.wrap(title, 68, initial_indent='  ', subsequent_indent='    '):
            print(l)
        print('  {0} - {1} - {2}'.format(issue['fields']['issuetype']['name'], issue['fields']['priority']['name'], 'https://tracker.moodle.org/browse/' + issue['key']))
        status = '{0} {1} {2}'.format(issue['fields']['status']['name'], resolution, resolutiondate).strip()
        print('  {0}'.format(status))

        print('-' * 72)
        components = '{0}: {1}'.format('Components', ', '.join([c['name'] for c in issue['fields']['components']]))
        for l in textwrap.wrap(components, 68, initial_indent='  ', subsequent_indent='              '):
            print(l)
        if issue['fields']['labels']:
            labels = '{0}: {1}'.format('Labels', ', '.join(issue['fields']['labels']))
            for l in textwrap.wrap(labels, 68, initial_indent='  ', subsequent_indent='          '):
                print(l)

        vw = '[ V: %d - W: %d ]' % (issue['fields']['votes']['votes'], issue['fields']['watches']['watchCount'])
        print('{0:->70}--'.format(vw))
        print('{0:<20}: {1} ({2}) on {3}'.format('Reporter', issue['fields']['reporter']['displayName'], issue['fields']['reporter']['name'], created))

        if issue['fields'].get('assignee') != None:
            print('{0:<20}: {1} ({2})'.format('Assignee', issue['fields']['assignee']['displayName'], issue['fields']['assignee']['name']))
        if issue['named'].get('Peer reviewer'):
            print('{0:<20}: {1} ({2})'.format('Peer reviewer', issue['named']['Peer reviewer']['displayName'], issue['named']['Peer reviewer']['name']))
        if issue['named'].get('Integrator'):
            print('{0:<20}: {1} ({2})'.format('Integrator', issue['named']['Integrator']['displayName'], issue['named']['Integrator']['name']))
        if issue['named'].get('Tester'):
            print('{0:<20}: {1} ({2})'.format('Tester', issue['named']['Tester']['displayName'], issue['named']['Tester']['name']))

        if args.testing and issue['named'].get('Testing Instructions'):
            print('-' * 72)
            print('Testing instructions:')
            for l in issue['named'].get('Testing Instructions').split('\r\n'):
                print('  ' + l)

        print('-' * 72)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
from ..tools import yesOrNo
from ..command import Command


class UninstallCommand(Command):

    _description = 'Uninstall a Moodle instance'
    _arguments = [
        (
            ['name'],
            {
                'default': None,
                'help': 'name of the instance',
                'metavar': 'name',
                'nargs': '?'
            }
        ),
        (
            ['-y'],
            {
                'action': 'store_true',
                'dest': 'do',
                'help': 'do not ask for confirmation'
            }
        )
    ]

    def run(self, args):

        M = self.Wp.resolve(args.name)
        if not M:
            raise Exception('This is not a Moodle instance')
        elif not M.isInstalled():
            logging.info('This instance is not installed')
            return

        if not args.do:
            if not yesOrNo('Are you sure?'):
                logging.info('Aborting...')
                return

        logging.info('Uninstalling %s...' % M.get('identifier'))
        M.uninstall()
        logging.info('Done.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2012 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
import re
import shlex
import subprocess


class Git(object):

    _path = None
    _bin = None

    def __init__(self, path, bin='/usr/bin/git'):
        self.setPath(path)
        self.setBin(bin)

    def add(self, path):
        """Add a file/path"""
        cmd = 'add %s' % (path)
        result = self.execute(cmd)
        return result[0] == 0

    def addRemote(self, name, remote):
        cmd = 'remote add %s %s' % (name, remote)
        result = self.execute(cmd)
        return result[0] == 0

    def apply(self, files):
        if type(files) == list:
            files = ' '.join(files)
        cmd = 'apply --check %s' % (files)
        result = self.execute(cmd)
        if result[0] != 0:
            return False
        cmd = 'am %s' % (files)
        result = self.execute(cmd)
        return result[0] == 0

    def checkout(self, branch):
        if self.currentBranch == branch:
            return True
        cmd = 'checkout %s' % branch
        result = self.execute(cmd)
        return result[0] == 0

    def commit(self, filepath=None):
        """Wrapper for the commit command"""
        cmd = 'commit'
        if filepath:
            cmd += ' -F %s' % (filepath)
        result = self.execute(cmd)
        return result[0] == 0

    def conflictingFiles(self):
        cmd = 'diff --name-only --diff-filter=U'
        result = self.execute(cmd)
        if result[0] == 0:
            return result[1].strip().split('\n')
        else:
            return False

    def createBranch(self, branch, track=None):
        if track != None:
            cmd = 'branch --track %s %s' % (branch, track)
        else:
            cmd = 'branch %s' % branch

        result = self.execute(cmd)
        return result[0] == 0

    def createPatch(self, ref, saveTo=None):
        cmd = 'format-patch %s --stdout' % (ref)
        result = self.execute(cmd)
        if result[0] != 0:
            return False
        elif saveTo != None:
            with open(saveTo, 'w') as f:
                f.write(result[1])
                return True
            return False
        return result[1]

    def currentBranch(self):
        cmd = 'symbolic-ref -q HEAD'
        result = self.execute(cmd)
        if result[0] != 0:
            return 'HEAD'
        else:
            return result[1].replace('refs/heads/', '').strip()

    def delRemote(self, remote):
        cmd = 'remote rm %s' % remote
        result = self.execute(cmd)
        return result[0] == 0

    def execute(self, cmd, path=None):
        if path == None:
            path = self.getPath()

        if not self.isRepository(path):
            raise Exception('This is not a Git repository')

        if not type(cmd) == list:
            cmd = shlex.split(str(cmd))
        cmd.insert(0, self.getBin())

        logging.debug(' '.join(cmd))

        try:
            proc = subprocess.Popen(cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=path
            )
            (stdout, stderr) = proc.communicate()
        except KeyboardInterrupt as e:
            proc.kill()
            raise e

        return (proc.returncode, stdout.decode('utf-8'), stderr.decode('utf-8'))

    def fetch(self, remote='', ref=''):
        remote = self.updateUnauthenticatedGithub(remote)
        cmd = 'fetch %s %s' % (remote, ref)
        result = self.execute(cmd)
        return result[0] == 0

    def getConfig(self, name):
        cmd = 'config --get %s' % name
        result = self.execute(cmd)
        if result[0] == 0:
            return result[1].strip()
        else:
            return None

    def getRemote(self, remote):
        remotes = self.getRemotes()
        return remotes.get(remote, None)

    def getRemotes(self):
        """Return the remotes"""
        cmd = 'remote -v'
        result = self.execute(cmd)
        remotes = None
        if result[0] == 0:
            remotes = {}
            for line in result[1].split('\n'):
                if not line: continue
                (remote, repo) = re.split('\s+', line, 1)
                repo = re.sub(' \(push\)|\(fetch\)$', '', repo)
                remotes[remote] = repo
        return remotes

    def hasBranch(self, branch, remote=''):
        if remote != '':
            cmd = 'show-ref --verify --quiet "refs/remotes/%s/%s"' % (remote, branch)
        else:
            cmd = 'show-ref --verify --quiet "refs/heads/%s"' % branch
        (returncode, stdout, stderr) = self.execute(cmd)
        return returncode == 0

    def hashes(self, ref='', format='%H', limit=30):
        """Returns the latest hashes from git log"""
        hashlist = self.log(count=limit, format=format, since=ref)
        return hashlist.split('\n')[:-1]

    def isRepository(self, path=None):
        if path == None:
            path = self.getPath()

        cmd = shlex.split(str('%s log -1') % self.getBin())
        proc = subprocess.Popen(cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=path
        )
        proc.wait()
        return proc.returncode == 0

    def updateUnauthenticatedGithub(self, path):
        """Detect and fix unauthenticated github URLs"""
        # https://github.blog/2021-09-01-improving-git-protocol-security-github/#no-more-unauthenticated-git
        if "git://github.com/" in path:
            logging.warn("Repository at {} uses unauthenticated github which is no longer supported. Converting to https.".format(path))
            return path.replace('git://github.com', 'https://github.com')
        return path

    def log(self, count=10, since=None, path=None, format=None):
        """Calls the log command and returns the raw output"""
        cmd = 'log'
        if count != None and count != 0:
            cmd += ' -n %d ' % (int(count))
        if format != None:
            cmd += ' --format=%s ' % (format)
        if since != None:
            cmd += ' %s ' % (since)
        if path != None:
            cmd += ' -- %s' % (path)

        (returncode, stdout, stderr) = self.execute(cmd)
        if returncode != 0:
            raise GitException('Error calling git log. Command: %s' % (cmd))

        return stdout

    def merge(self, ref):
        """Wrapper for the merge command"""
        cmd = 'merge %s' % (ref)
        (returncode, stdout, stderr) = self.execute(cmd)
        return returncode == 0

    def messages(self, count=10, since=None, path=None):
        """Return the latest titles of the commit messages"""
        messages = self.log(count=count, since=since, path=path, format='%s')
        return messages.split('\n')[:-1]

    def pick(self, refs=None, abort=None, continu=None):
        """Wrapper for the cherry-pick command

        The typo in the argument 'continu' is voluntarily there, continue is a reserved word.
        """
        args = ''
        if refs != None:
            if type(refs) == list:
                refs = ' '.join(refs)
            args = refs
        elif abort != None:
            args = '--abort'
        elif continu != None:
            args = '--continue'
        cmd = 'cherry-pick %s' % (args)
        return self.execute(cmd)

    def pull(self, remote='', ref=''):
        cmd = 'pull %s %s' % (remote, ref)
        return self.execute(cmd)

    def push(self, remote='', branch='', force=None):
        if force:
            force = '--force '
        else:
            force = ''
        cmd = 'push %s%s %s' % (force, remote, branch)
        return self.execute(cmd)

    def rebase(self, base=None, branch=None, abort=False):
        cmd = None
        if abort:
            cmd = 'rebase --abort'
        elif base != None and branch != None:
            # Rebase automatically checks out the branch before rebasing
            cmd = 'rebase %s %s' % (base, branch)
        if cmd == None:
            raise Exception('Missing arguments for calling rebase')
        return self.execute(cmd)

    def remoteBranches(self, remote):
        pattern = 'refs/remotes/%s' % remote
        cmd = 'show-ref'

        result = self.execute(cmd)
        if result[0] != 0:
            return []

        refs = []
        for ref in result[1].split('\n'):
            try:
                (hash, ref) = ref.split(' ', 1)
            except ValueError:
                continue
            if not ref.startswith(pattern):
                continue
            ref = ref.replace(pattern, '').strip('/')
            if ref == 'HEAD':
                continue
            refs.append([hash, ref])
        return refs

    def reset(self, to, hard=False):
        mode = ''
        if hard:
            mode = '--hard'
        cmd = 'reset %s %s' % (mode, to)
        result = self.execute(cmd)
        return result[0] == 0

    def setConfig(self, name, value):
        cmd = 'config %s %s' % (name, value)
        result = self.execute(cmd)
        return result[0] == 0

    def setRemote(self, remote, url):
        if not self.getRemote(remote):
            cmd = 'remote add %s %s' % (remote, url)
        else:
            cmd = 'remote set-url %s %s' % (remote, url)
        result = self.execute(cmd)
        return result[0] == 0

    def stash(self, command='save', untracked=False):
        cmd = 'stash %s' % command
        if untracked:
            cmd += ' --include-untracked'
        return self.execute(cmd)

    def status(self):
        return self.execute('status')

    def getBin(self):
        return self._bin

    def setBin(self, bin):
        self._bin = str(bin)

    def getPath(self):
        return self._path

    def setPath(self, path):
        self._path = str(path)


class GitException(Exception):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import os
import re
import shutil
import subprocess
from .. import git
from ..command import Command
from ..tools import mkdir, resolveEditor


class DoctorCommand(Command):

    _arguments = [
        (
            ['--fix'],
            {
                'action': 'store_true',
                'help': 'Automatically fix all the identified problems'
            }
        ),
        (
            ['--all'],
            {
                'action': 'store_true',
                'help': 'Enable all the checks'
            }
        ),
        (
            ['--branch'],
            {
                'action': 'store_true',
                'help': 'Check the branch checked out on your integration instances'
            }
        ),
        (
            ['--cached'],
            {
                'action': 'store_true',
                'help': 'Check the cached repositories'
            }
        ),
        (
            ['--dependencies'],
            {
                'action': 'store_true',
                'help': 'Check various dependencies'
            }
        ),
        (
            ['--directories'],
            {
                'action': 'store_true',
                'help': 'Check the directories set in the config file'
            }
        ),
        (
            ['--hi'],
            {
                'action': 'store_true',
                'help': 'What you see it totally unrelated to what you get',
                'silent': True
            }
        ),
        (
            ['--masterbranch'],
            {
                'action': 'store_true',
                'help': 'Check the status of the master branch'
            }
        ),
        (
            ['--remotes'],
            {
                'action': 'store_true',
                'help': 'Check the remotes of your instances'
            }
        ),
        (
            ['--symlink'],
            {
                'action': 'store_true',
                'help': 'Check the symlinks of the instances'
            }
        ),
        (
            ['--wwwroot'],
            {
                'action': 'store_true',
                'help': 'Check the $CFG->wwwroot of your instances'
            }
        )
    ]
    _description = 'Perform several checks on your current installation'

    def run(self, args):

        optionsCount = sum([1 for k, v in list(vars(args).items()) if v != False])
        if optionsCount == 0 or (optionsCount == 1 and args.fix):
            self.argumentError('You should probably tell me what symptoms you are experiencing')

        allChecks = False
        if args.all:
            allChecks = True

        # Check directories
        if args.directories or allChecks:
            self.directories(args)

        # Check the cached remotes
        if args.cached or allChecks:
            self.cachedRepositories(args)

        # Check the dependencies
        if args.dependencies or allChecks:
            self.dependencies(args)

        # Check instances remotes
        if args.remotes or allChecks:
            self.remotes(args)

        # Check instances wwwroot
        if args.wwwroot or allChecks:
            self.wwwroot(args)

        # Check instances symlink
        if args.symlink or allChecks:
            self.symlink(args)

        # Check the branches
        if args.branch or allChecks:
            self.branch(args)

        # Check the master branch
        if args.masterbranch or allChecks:
            self.masterbranch(args)

        # Check what you see is what you get
        if args.hi:
            self.hi(args)

    def branch(self, args):
        """Make sure the correct branch is checked out. Only on integration branches."""

        print('Checking integration instances branches')

        if not self._checkWorkplace():
            return

        instances = self.Wp.list(integration=True)

        for identifier in instances:
            M = self.Wp.get(identifier)
            stablebranch = M.get('stablebranch')
            currentbranch = M.currentBranch()
            if stablebranch != currentbranch:
                print('  %s is on branch %s instead of %s' % (identifier, currentbranch, stablebranch))
                if args.fix:
                    print('    Checking out %s' % (stablebranch))
                    if not M.git().checkout(stablebranch):
                        print('      Error: Checkout unsucessful!')

    def cachedRepositories(self, args):
        """Ensure that the cached repositories are valid"""

        print('Checking cached repositories')
        cache = os.path.abspath(os.path.realpath(os.path.expanduser(self.C.get('dirs.mdk'))))

        dirs = [
            {
                'dir': os.path.join(cache, 'moodle.git'),
                'url': self.C.get('remotes.stable')
            },
            {
                'dir': os.path.join(cache, 'integration.git'),
                'url': self.C.get('remotes.integration')
            },
        ]

        for d in dirs:
            directory = d['dir']
            name = os.path.split(directory)[1]

            if os.path.isdir(directory):
                if os.path.isdir(os.path.join(directory, '.git')):
                    print('  %s is not a bare repository' % name)
                    if args.fix:
                        print('    Renaming %s/.git directory to %s' % (directory, directory))
                        os.rename(directory, directory + '.tmp')
                        os.rename(os.path.join(directory + '.tmp', '.git'), directory)
                        shutil.rmtree(directory + '.tmp')

                repo = git.Git(directory, self.C.get('git'))
                if repo.getConfig('core.bare') != 'true':
                    print('  %s core.bare is not set to true' % name)
                    if args.fix:
                        print('    Setting core.bare to true')
                        repo.setConfig('core.bare', 'true')

                if repo.getConfig('remote.origin.url') != d['url']:
                    print('  %s uses an different origin (%s)' % (name, repo.getConfig('remote.origin.url')))
                    if args.fix:
                        print('    Setting remote.origin.url to %s' % d['url'])
                        repo.setConfig('remote.origin.url', d['url'])

                if repo.getConfig('remote.origin.fetch') != '+refs/*:refs/*':
                    print('  %s fetch value is invalid (%s)' % (name, repo.getConfig('remote.origin.fetch')))
                    if args.fix:
                        print('    Setting remote.origin.fetch to %s' % '+refs/*:refs/*')
                        repo.setConfig('remote.origin.fetch', '+refs/*:refs/*')

    def dependencies(self, args):
        """Check that various dependencies are met"""

        print('Checking dependencies')

        # Check binaries.
        hasErrors = False
        for k in ['git', 'php', 'java', 'recess', 'grunt', 'lessc', 'shifter', 'yuidoc']:
            path = self.C.get(k)
            if not path or not os.path.isfile(path):
                print('  The path to \'%s\' is invalid: %s' % (k, path))
                hasErrors = True
        if hasErrors and args.fix:
            print('    Please manually fix the paths in your config file')

        # Checking editor.
        editor = resolveEditor()
        if editor:
            try:
                # Check if it is callable.
                proc = subprocess.Popen(editor, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                proc.kill()
            except OSError:
                editor = None

        if not editor:
            print('  Could not resolve the path to your editor')
            if args.fix:
                print('    Set $EDITOR, /usr/bin/editor, or use: mdk config set editor [path]')

    def directories(self, args):
        """Check that the directories are valid"""

        print('Checking directories')
        for k, d in list(self.C.get('dirs').items()):
            d = os.path.abspath(os.path.realpath(os.path.expanduser(d)))
            if not os.path.isdir(d):
                print('  %s does not exist' % d)
                if args.fix:
                    print('    Creating %s' % d)
                    mkdir(d, 0o777)

        if not self._checkWorkplace():
            return

        # Checking extra directory.
        instances = self.Wp.list()
        for identifier in instances:
            d = self.Wp.getPath(identifier, 'extra')
            if not os.path.isdir(d):
                print('  %s does not exist' % d)
                if args.fix:
                    print('    Creating %s' % d)
                    mkdir(d, 0o777)

    def masterbranch(self, args):
        """Checks the current master branch and the value set in config."""

        print('Checking master branch')

        if not self._checkWorkplace():
            return

        repoPath = self.Wp.getCachedRemote()
        if not os.path.isdir(repoPath):
            return

        try:
            self.Wp.updateCachedClones(verbose=False)
        except Exception:
            print('  Could not update clone, please try again.')
            return

        repo = git.Git(repoPath, self.C.get('git'))
        result = repo.execute(['show', 'master:version.php'])
        if result[0] != 0:
            print('  Could not read the master version.php')
            return

        reBranch = re.compile(r'^\s*\$branch\s*=\s*(?P<brackets>[\'"])?([0-9]+)(?P=brackets)\s*;')
        latestBranch = None
        for line in result[1].split('\n'):
            if reBranch.search(line):
                latestBranch = int(reBranch.search(line).group(2))

        masterBranch = int(self.C.get('masterBranch'))
        if not latestBranch:
            print('  Oops, could not identify the mater branch')
        elif masterBranch != latestBranch:
            print('  The config masterBranch is set to %d, expecting %d' % (masterBranch, latestBranch))
            if args.fix:
                print('    Setting masterBranch to %d' % (latestBranch))
                self.C.set('masterBranch', latestBranch)


    def hi(self, args):
        """I wonder what is the purpose of this...

            hint #1: 1341
            hint #2: dobedobedoh
        """

        if args.fix:
            print('The horse is a noble animal')
        else:
            print('<em>Hi</em>')

    def symlink(self, args):
        """Check that the symlinks exist"""

        print('Checking symlinks')

        if not self._checkWorkplace():
            return

        instances = self.Wp.list()
        for identifier in instances:

            wwwLink = os.path.join(self.Wp.www, identifier)
            if not os.path.exists(wwwLink):
                print('  Missing link to www for %s' % (identifier))
                if args.fix:
                    print('    Creating www symlink for %s' % (identifier))
                    os.symlink(self.Wp.getPath(identifier, 'www'), wwwLink)

            extraLink = os.path.join(self.Wp.getMdkWebDir(), identifier)
            if not os.path.exists(extraLink):
                print('  Missing link to extra for %s' % (identifier))
                if args.fix:
                    print('    Creating extra symlink for %s' % (identifier))
                    os.symlink(self.Wp.getPath(identifier, 'extra'), extraLink)


    def remotes(self, args):
        """Check that the correct remotes are used"""

        print('Checking remotes')

        if not self._checkWorkplace():
            return

        remotes = {
            'mine': self.C.get('remotes.mine'),
            'stable': self.Wp.getCachedRemote() if self.C.get('useCacheAsUpstreamRemote') else self.C.get('remotes.stable'),
            'integration': self.Wp.getCachedRemote(True) if self.C.get('useCacheAsUpstreamRemote') else self.C.get('remotes.integration')
        }
        myRemote = self.C.get('myRemote')
        upstreamRemote = self.C.get('upstreamRemote')

        instances = self.Wp.list()
        for identifier in instances:
            M = self.Wp.get(identifier)

            remote = M.git().getRemote(myRemote)
            if remote != remotes['mine']:
                print('  %s: Remote %s is %s, not %s' % (identifier, myRemote, remote, remotes['mine']))
                if (args.fix):
                    print('    Setting %s to %s' % (myRemote, remotes['mine']))
                    M.git().setRemote(myRemote, remotes['mine'])

            expected = remotes['stable'] if M.isStable() else remotes['integration']
            remote = M.git().getRemote(upstreamRemote)
            if remote != expected:
                print('  %s: Remote %s is %s, not %s' % (identifier, upstreamRemote, remote, expected))
                if (args.fix):
                    print('    Setting %s to %s' % (upstreamRemote, expected))
                    M.git().setRemote(upstreamRemote, expected)

    def wwwroot(self, args):
        """Check the wwwroot of the instances"""

        print('Checking wwwroot')

        if not self._checkWorkplace():
            return

        instances = self.Wp.resolveMultiple(self.Wp.list())


        for M in instances:
            if not M.isInstalled():
                continue
            else:
                actual = M.get('wwwroot')
                expected = self.Wp.getUrl(M.get('identifier'))
                if actual != expected:
                    print('  %s: Found %s, not %s' % (M.get('identifier'), actual, expected))
                    if args.fix:
                        print('    Setting %s on %s' % (expected, M.get('identifier')))
                        M.updateConfig('wwwroot', expected)

    def _checkWorkplace(self, indent=2):
        """Returns whether the workplace is available or, and print a message if it is not."""
        try:
            self.Wp
        except ImportError:
            print(' ' * indent + 'The workplace could not be loaded, did you install the dependencies?')
            return False
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import sys
import logging
from ..command import Command
from ..exceptions import UpgradeNotAllowed

class UpgradeCommand(Command):

    _arguments = [
        (
            ['-a', '--all'],
            {
                'action': 'store_true',
                'dest': 'all',
                'help': 'upgrade each instance'
            }
        ),
        (
            ['-i', '--integration'],
            {
                'action': 'store_true',
                'dest': 'integration',
                'help': 'upgrade integration instances'
            }
        ),
        (
            ['-s', '--stable'],
            {
                'action': 'store_true',
                'dest': 'stable',
                'help': 'upgrade stable instances'
            }
        ),
        (
            ['-n', '--no-checkout'],
            {
                'action': 'store_true',
                'dest': 'nocheckout',
                'help': 'do not checkout the stable branch before upgrading'
            }
        ),
        (
            ['-u', '--update'],
            {
                'action': 'store_true',
                'dest': 'update',
                'help': 'update the instance before running the upgrade script'
            }
        ),
        (
            ['names'],
            {
                'default': None,
                'help': 'name of the instances',
                'metavar': 'names',
                'nargs': '*'
            }
        )
    ]
    _description = 'Run the Moodle upgrade script'

    def run(self, args):

        names = args.names
        if args.all:
            names = self.Wp.list()
        elif args.integration or args.stable:
            names = self.Wp.list(integration=args.integration, stable=args.stable)

        Mlist = self.Wp.resolveMultiple(names)
        if len(Mlist) < 1:
            raise Exception('No instances to work on. Exiting...')

        # Updating cache if required
        if args.update:
            print('Updating cached repositories')
            self.Wp.updateCachedClones(verbose=False)

        errors = []

        for M in Mlist:
            if args.update:
                logging.info('Updating %s...' % M.get('identifier'))
                try:
                    M.update()
                except Exception as e:
                    errors.append(M)
                    logging.warning('Error during update. Skipping...')
                    logging.debug(e)
                    continue
            logging.info('Upgrading %s...' % M.get('identifier'))

            try:
                M.upgrade(args.nocheckout)
            except UpgradeNotAllowed as e:
                logging.info('Skipping upgrade of %s (not allowed)' % (M.get('identifier')))
                logging.debug(e)
            except Exception as e:
                errors.append(M)
                logging.warning('Error during the upgrade of %s' % M.get('identifier'))
                logging.debug(e)
            logging.info('')
        logging.info('Done.')

        if errors and len(Mlist) > 1:
            logging.info('')
            logging.warning('/!\ Some errors occurred on the following instances:')
            for M in errors:
                logging.warning('- %s' % M.get('identifier'))
            # TODO Do not use sys.exit() but handle error code
            sys.exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import argparse
from ..command import Command


class AliasCommand(Command):

    _arguments = [
        (
            ['action'],
            {
                'metavar': 'action',
                'help': 'the action to perform',
                'sub-commands':
                    {
                        'list': (
                            {
                                'help': 'list the aliases'
                            },
                            []
                        ),
                        'show': (
                            {
                                'help': 'display an alias'
                            },
                            [
                                (
                                    ['alias'],
                                    {
                                        'type': str,
                                        'metavar': 'alias',
                                        'default': None,
                                        'help': 'alias to display'
                                    }
                                )
                            ]
                        ),
                        'add': (
                            {
                                'help': 'adds an alias'
                            },
                            [
                                (
                                    ['alias'],
                                    {
                                        'type': str,
                                        'metavar': 'alias',
                                        'default': None,
                                        'help': 'alias name'
                                    }
                                ),
                                (
                                    ['definition'],
                                    {
                                        'type': str,
                                        'metavar': 'command',
                                        'default': None,
                                        'nargs': argparse.REMAINDER,
                                        'help': 'alias definition'
                                    }
                                )
                            ]
                        ),
                        'remove': (
                            {
                                'help': 'remove an alias'
                            },
                            [
                                (
                                    ['alias'],
                                    {
                                        'type': str,
                                        'metavar': 'alias',
                                        'default': None,
                                        'help': 'alias to remove'
                                    }
                                )
                            ]
                        ),
                        'set': (
                            {
                                'help': 'update/add an alias'
                            },
                            [
                                (
                                    ['alias'],
                                    {
                                        'type': str,
                                        'metavar': 'alias',
                                        'default': None,
                                        'help': 'alias name'
                                    }
                                ),
                                (
                                    ['definition'],
                                    {
                                        'type': str,
                                        'metavar': 'command',
                                        'default': None,
                                        'nargs': argparse.REMAINDER,
                                        'help': 'alias definition'
                                    }
                                )
                            ]
                        )
                    }
            }
        )
    ]
    _description = 'Manage your aliases'

    def run(self, args):
        if args.action == 'list':
            aliases = self.C.get('aliases')
            for alias, command in list(aliases.items()):
                print('{0:<20}: {1}'.format(alias, command))

        elif args.action == 'show':
            alias = self.C.get('aliases.%s' % args.alias)
            if alias != None:
                print(alias)

        elif args.action == 'add':
            self.C.add('aliases.%s' % args.alias, ' '.join(args.definition))

        elif args.action == 'set':
            self.C.set('aliases.%s' % args.alias, ' '.join(args.definition))

        elif args.action == 'remove':
            self.C.remove('aliases.%s' % args.alias)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2014 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import os
from . import jira
import logging


class Fetch(object):
    """Holds the logic and processing to fetch a remote and following actions into an instance"""

    _M = None
    _ref = None
    _repo = None

    _canCreateBranch = True
    _hasstashed = False

    def __init__(self, M, repo=None, ref=None):
        self._M = M
        self._repo = repo
        self._ref = ref

    def checkout(self):
        """Fetch and checkout the fetched branch"""
        self.fetch()
        logging.info('Checking out branch as FETCH_HEAD')
        if not self.M.git().checkout('FETCH_HEAD'):
            raise FetchException('Could not checkout FETCH_HEAD')

    def fetch(self):
        """Perform the fetch"""
        if not self.repo:
            raise FetchException('The repository to fetch from is unknown')
        elif not self.ref:
            raise FetchException('The ref to fetch is unknown')

        git = self.M.git()
        logging.info('Fetching %s from %s' % (self.ref, self.repo))
        result = git.fetch(remote=self.repo, ref=self.ref)
        if not result:
            raise FetchException('Error while fetching %s from %s' % (self.ref, self.repo))

    def _merge(self):
        """Protected method to merge FETCH_HEAD into the current branch"""
        logging.info('Merging into current branch')
        if not self.M.git().merge('FETCH_HEAD'):
            raise FetchException('Merge failed, resolve the conflicts and commit')

    def pull(self, into=None, track=None):
        """Fetch and merge the fetched branch into a branch passed as param"""
        self._stash()

        try:
            self.fetch()
            git = self.M.git()

            if into:
                logging.info('Switching to branch %s' % (into))

                if not git.hasBranch(into):
                    if self.canCreateBranch:
                        if not git.createBranch(into, track=track):
                            raise FetchException('Could not create the branch %s' % (into))
                    else:
                        raise FetchException('Branch %s does not exist and create branch is forbidden' % (into))

                if not git.checkout(into):
                    raise FetchException('Could not checkout branch %s' % (into))

            self._merge()

        except FetchException as e:
            if self._hasstashed:
                logging.warning('An error occurred. Some files may have been left in your stash.')
            raise e

        self._unstash()

    def setRef(self, ref):
        """Set the reference to fetch"""
        self._ref = ref

    def setRepo(self, repo):
        """Set the repository to fetch from"""
        self._repo = repo

    def _stash(self):
        """Protected method to stash"""
        stash = self.M.git().stash(untracked=True)
        if stash[0] != 0:
            raise FetchException('Error while trying to stash your changes')
        elif not stash[1].startswith('No local changes'):
            logging.info('Stashed your local changes')
            self._hasstashed = True

    def _unstash(self):
        """Protected method to unstash"""
        if self._hasstashed:
            pop = self.M.git().stash(command='pop')
            if pop[0] != 0:
                logging.error('An error occurred while unstashing your changes')
            else:
                logging.info('Popped the stash')
        self._hasstashed = False

    @property
    def canCreateBranch(self):
        return self._canCreateBranch

    @property
    def into(self):
        return self._into

    @property
    def M(self):
        return self._M

    @property
    def ref(self):
        return self._ref

    @property
    def repo(self):
        return self._repo


class FetchTracker(Fetch):
    """Pretty dodgy implementation of Fetch to work with the tracker.

        If a list of patches is set, we override the git methods to fetch from a remote
        to use the patches instead. I am not super convinced by this design, but at
        least the logic to fetch/pull/merge is more or less self contained.
    """

    _J = None
    _cache = None
    _patches = None

    def __init__(self, *args, **kwargs):
        super(FetchTracker, self).__init__(*args, **kwargs)
        self._J = jira.Jira()
        self._cache = {}

    def checkout(self):
        if not self.patches:
            return super(FetchTracker, self).checkout()

        self.fetch()

    def fetch(self):
        if not self.patches:
            return super(FetchTracker, self).fetch()

        for patch in self.patches:
            j = 0
            dest = None
            while True:
                downloadedTo = patch.get('filename') + (('.' + str(j)) if j > 0 else '')
                dest = os.path.join(self.M.get('path'), downloadedTo)
                j += 1
                if not os.path.isfile(dest):
                    patch['downloadedTo'] = downloadedTo
                    break
            
            logging.info('Downloading patch as %s' % (patch.get('downloadedTo')))
            if not dest or not self.J.download(patch.get('url'), dest):
                raise FetchTrackerException('Failed to download the patch to %s' % (dest))


    def _merge(self):
        if not self.patches:
            return super(FetchTracker, self)._merge()

        patchList = [patch.get('downloadedTo') for patch in self.patches]
        git = self.M.git()
        if not git.apply(patchList):
            raise FetchTrackerException('Could not apply the patch(es), please apply manually')
        else:
            for f in patchList:
                os.remove(f)
        logging.info('Patches applied successfully')

    def getPullInfo(self, mdl):
        """Return the pull information

            This implements its own local cache because we could potentially
            call it multiple times during the same request. This is bad though.
        """
        if mdl not in self._cache:
            issueInfo = self.J.getPullInfo(mdl)
            self._cache[mdl] = issueInfo
        return self._cache[mdl]

    def setFromTracker(self, mdl, branch):
        """Sets the repo and ref according to the tracker information"""
        issueInfo = self.getPullInfo(mdl)

        repo = issueInfo.get('repo', None)
        if not repo:
            raise FetchTrackerRepoException('Missing information about the repository to pull from on %s' % (mdl))

        ref = issueInfo.get('branches').get(str(branch), None)
        if not ref:
            raise FetchTrackerBranchException('Could not find branch info on %s' % (str(branch), mdl))

        self.setRepo(repo)
        self.setRef(ref.get('branch'))

    def usePatches(self, patches):
        """List of patches (returned by jira.Jira.getAttachments) to work with instead of the standard repo and ref"""
        self._patches = patches

    @property
    def J(self):
        return self._J

    @property
    def patches(self):
        return self._patches


class FetchException(Exception):
    pass


class FetchTrackerException(FetchException):
    pass


class FetchTrackerBranchException(FetchTrackerException):
    pass


class FetchTrackerRepoException(FetchTrackerException):
    pass
